@@ -13,6 +13,7 @@
 
 #include <Python.h>
 #include <assert.h>
+#include <err.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -38,8 +39,8 @@
 #include <arpa/inet.h>
 
 
-#include "netbsd.h"
 #include "netbsd_socks.h"
+#include "netbsd.h"
 #include "../../_psutil_common.h"
 
 #define PSUTIL_KPT2DOUBLE(t) (t ## _sec + t ## _usec / 1000000.0)
@@ -51,16 +52,6 @@
 // ============================================================================
 
 
-int
-psutil_raise_ad_or_nsp(long pid) {
-    // Set exception to AccessDenied if pid exists else NoSuchProcess.
-    if (psutil_pid_exists(pid) == 0)
-        NoSuchProcess();
-    else
-        AccessDenied();
-}
-
-
 int
 psutil_kinfo_proc(pid_t pid, kinfo_proc *proc) {
     // Fills a kinfo_proc struct based on process pid.
@@ -123,31 +114,6 @@ kinfo_getfile(pid_t pid, int* cnt) {
 }
 
 
-int
-psutil_pid_exists(pid_t pid) {
-    // Return 1 if PID exists in the current process list, else 0, -1
-    // on error.
-    // TODO: this should live in _psutil_posix.c but for some reason if I
-    // move it there I get a "include undefined symbol" error.
-    int ret;
-    if (pid < 0)
-        return 0;
-    ret = kill(pid , 0);
-    if (ret == 0)
-        return 1;
-    else {
-        if (ret == ESRCH)
-            return 0;
-        else if (ret == EPERM)
-            return 1;
-        else {
-            PyErr_SetFromErrno(PyExc_OSError);
-            return -1;
-        }
-    }
-}
-
-
 // XXX: This is no longer used as per
 // https://github.com/giampaolo/psutil/pull/557#issuecomment-171912820
 // Current implementation uses /proc instead.
@@ -195,7 +161,13 @@ psutil_proc_exe(PyObject *self, PyObject *args) {
         else
             strcpy(pathname, "");
     }
+
+#if PY_MAJOR_VERSION >= 3
+    return PyUnicode_DecodeFSDefault(pathname);
+#else
     return Py_BuildValue("s", pathname);
+#endif
+
 #else
     return Py_BuildValue("s", "");
 #endif
@@ -316,13 +288,14 @@ psutil_get_proc_list(kinfo_proc **procList, size_t *procCount) {
     kd = kvm_openfiles(NULL, NULL, NULL, KVM_NO_FILES, errbuf);
 
     if (kd == NULL) {
-        PyErr_Format(PyExc_RuntimeError, "kvm_openfiles() failed: %s", errbuf);
+        PyErr_Format(
+            PyExc_RuntimeError, "kvm_openfiles() syscall failed: %s", errbuf);
         return errno;
     }
 
     result = kvm_getproc2(kd, KERN_PROC_ALL, 0, sizeof(kinfo_proc), &cnt);
     if (result == NULL) {
-        PyErr_Format(PyExc_RuntimeError, "kvm_getproc2() failed");
+        PyErr_Format(PyExc_RuntimeError, "kvm_getproc2() syscall failed");
         kvm_close(kd);
         return errno;
     }
@@ -412,7 +385,11 @@ psutil_get_cmdline(pid_t pid) {
     // separator
     if (argsize > 0) {
         while (pos < argsize) {
+#if PY_MAJOR_VERSION >= 3
+            py_arg = PyUnicode_DecodeFSDefault(&argstr[pos]);
+#else
             py_arg = Py_BuildValue("s", &argstr[pos]);
+#endif
             if (!py_arg)
                 goto error;
             if (PyList_Append(py_retlist, py_arg))
@@ -434,37 +411,32 @@ error:
 }
 
 
+/*
+ * Virtual memory stats, taken from:
+ * https://github.com/satterly/zabbix-stats/blob/master/src/libs/zbxsysinfo/
+ *     netbsd/memory.c
+ */
 PyObject *
 psutil_virtual_mem(PyObject *self, PyObject *args) {
-    int64_t total_physmem;
     size_t size;
     struct uvmexp_sysctl uv;
-    int physmem_mib[] = {CTL_HW, HW_PHYSMEM64};
-    int uvmexp_mib[] = {CTL_VM, VM_UVMEXP2};
+    int mib[] = {CTL_VM, VM_UVMEXP2};
     long pagesize = getpagesize();
 
-    size = sizeof(total_physmem);
-    if (sysctl(physmem_mib, 2, &total_physmem, &size, NULL, 0) < 0) {
-        PyErr_SetFromErrno(PyExc_OSError);
-        return NULL;
-    }
-
     size = sizeof(uv);
-    if (sysctl(uvmexp_mib, 2, &uv, &size, NULL, 0) < 0) {
+    if (sysctl(mib, 2, &uv, &size, NULL, 0) < 0) {
         PyErr_SetFromErrno(PyExc_OSError);
         return NULL;
     }
 
     return Py_BuildValue("KKKKKKKK",
-        (unsigned long long) total_physmem,  // total
-        (unsigned long long) uv.free * pagesize,  // free
-        (unsigned long long) uv.active * pagesize,  // active
-        (unsigned long long) uv.inactive * pagesize,  // inactive
-        (unsigned long long) uv.wired * pagesize,  // wired
-        // taken from:
-        // https://github.com/satterly/zabbix-stats/blob/master/src/libs/
-        //      zbxsysinfo/netbsd/memory.c
+        (unsigned long long) uv.npages << uv.pageshift,  // total
+        (unsigned long long) uv.free << uv.pageshift,  // free
+        (unsigned long long) uv.active << uv.pageshift,  // active
+        (unsigned long long) uv.inactive << uv.pageshift,  // inactive
+        (unsigned long long) uv.wired << uv.pageshift,  // wired
         (unsigned long long) uv.filepages + uv.execpages * pagesize,  // cached
+        // These are determined from /proc/meminfo in Python.
         (unsigned long long) 0,  // buffers
         (unsigned long long) 0  // shared
     );
@@ -498,8 +470,8 @@ psutil_swap_mem(PyObject *self, PyObject *args) {
     swap_total = swap_free = 0;
     for (i = 0; i < nswap; i++) {
         if (swdev[i].se_flags & SWF_ENABLE) {
-            swap_free += (swdev[i].se_nblks - swdev[i].se_inuse);
-            swap_total += swdev[i].se_nblks;
+            swap_total += swdev[i].se_nblks * DEV_BSIZE;
+            swap_free += (swdev[i].se_nblks - swdev[i].se_inuse) * DEV_BSIZE;
         }
     }
     free(swdev);
@@ -517,9 +489,9 @@ psutil_swap_mem(PyObject *self, PyObject *args) {
     }
 
     return Py_BuildValue("(LLLll)",
-                         swap_total * DEV_BSIZE,
-                         (swap_total - swap_free) * DEV_BSIZE,
-                         swap_free * DEV_BSIZE,
+                         swap_total,
+                         (swap_total - swap_free),
+                         swap_free,
                          (long) uv.pgswapin * pagesize,  // swap in
                          (long) uv.pgswapout * pagesize);  // swap out
 
@@ -538,9 +510,10 @@ psutil_proc_num_fds(PyObject *self, PyObject *args) {
     if (! PyArg_ParseTuple(args, "l", &pid))
         return NULL;
 
+    errno = 0;
     freep = kinfo_getfile(pid, &cnt);
     if (freep == NULL) {
-        psutil_raise_ad_or_nsp(pid);
+        psutil_raise_for_pid(pid, "kinfo_getfile() failed");
         return NULL;
     }
     free(freep);
@@ -551,6 +524,7 @@ psutil_proc_num_fds(PyObject *self, PyObject *args) {
 
 PyObject *
 psutil_per_cpu_times(PyObject *self, PyObject *args) {
+    // XXX: why static?
     static int maxcpus;
     int mib[3];
     int ncpu;
@@ -633,22 +607,19 @@ psutil_disk_io_counters(PyObject *self, PyObject *args) {
         PyErr_NoMemory();
         goto error;
     }
-    if (sysctl(mib, 2, stats, &len, NULL, 0) < 0 ) {
+    if (sysctl(mib, 3, stats, &len, NULL, 0) < 0 ) {
         PyErr_SetFromErrno(PyExc_OSError);
         goto error;
     }
 
     for (i = 0; i < dk_ndrive; i++) {
         py_disk_info = Py_BuildValue(
-            "(KKKKLL)",
+            "(KKKK)",
             stats[i].rxfer,
             stats[i].wxfer,
             stats[i].rbytes,
-            stats[i].wbytes,
-            // assume half read - half writes.
-            // TODO: why?
-            (long long) PSUTIL_KPT2DOUBLE(stats[i].time) / 2,
-            (long long) PSUTIL_KPT2DOUBLE(stats[i].time) / 2);
+            stats[i].wbytes
+        );
         if (!py_disk_info)
             goto error;
         if (PyDict_SetItemString(py_retdict, stats[i].name, py_disk_info))
@@ -666,3 +637,28 @@ error:
         free(stats);
     return NULL;
 }
+
+
+PyObject *
+psutil_cpu_stats(PyObject *self, PyObject *args) {
+    size_t size;
+    struct uvmexp_sysctl uv;
+    int uvmexp_mib[] = {CTL_VM, VM_UVMEXP2};
+
+    size = sizeof(uv);
+    if (sysctl(uvmexp_mib, 2, &uv, &size, NULL, 0) < 0) {
+        PyErr_SetFromErrno(PyExc_OSError);
+        return NULL;
+    }
+
+    return Py_BuildValue(
+        "IIIIIII",
+        uv.swtch,  // ctx switches
+        uv.intrs,  // interrupts - XXX always 0, will be determined via /proc
+        uv.softs,  // soft interrupts
+        uv.syscalls,  // syscalls - XXX always 0
+        uv.traps,  // traps
+        uv.faults,  // faults
+        uv.forks  // forks
+    );
+}
